# Copyright Red Hat 2017, Jake Hunsaker <jhunsake@redhat.com>
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import logging
import subprocess


class Cluster():

    def __init__(self, config, master):
        self.master = master
        self.config = config
        self.cluster_type = self.__class__.__name__
        self.node_list = None
        sos_plugins = []
        self.logger = logging.getLogger('clustersos')
        self.console = logging.getLogger('clustersos_console')
        if not getattr(self, 'packages', False):
            self.packages = ('',)
        if not getattr(self, 'sos_options', False):
            self.sos_options = {}
        if not getattr(self, "option_list", False):
            self.option_list = []
        self.options = []
        self._get_options()

    def _get_options(self):
        for opt in self.config['cluster_options']:
            if self.cluster_type in opt:
                self.options.append((opt.split('.')[1],
                                     self.config['cluster_options'][opt]
                                     )
                                    )

    def _fmt_msg(self, msg):
        return '[ %s ] %s' % (self.cluster_type, msg)

    def log_info(self, msg):
        '''Used to print info messages'''
        self.logger.info(self._fmt_msg(msg))
        self.console.info(msg)

    def log_error(self, msg):
        '''Used to print error messages'''
        self.logger.error(self._fmt_msg(msg))
        self.console.error(msg)

    def get_option(self, option):
        '''This is used to by clusters to check if a cluster option was
        supplied to clustersos.
        '''
        for opt in self.options:
            if option == opt[0]:
                return opt[1]
        return False

    def is_installed(self, pkg):
        cmd = self.master.host_facts['package_manager']['query'] + pkg
        res = self.exec_master_cmd(cmd)
        if res:
            return True
        return False

    def exec_master_cmd(self, cmd):
        '''Used to retrieve output from a (master) node in a cluster'''
        if self.config['need_sudo']:
            cmd = "sudo -S %s" % cmd
        if self.config['become_root']:
            cmd = "su -c '%s'" % cmd
        self.logger.debug('Running %s on %s' % (cmd, self.master.address))
        if self.master:
            sin, stdout, serr = self.master.client.exec_command(cmd,
                                                                get_pty=True
                                                                )
            if self.config['become_root']:
                sin.write(self.config['root_password'] + '\n')
                sin.flush()
            if self.config['need_sudo']:
                sin.write(self.config['sudo_pw'] + '\n')
                sin.flush()
            rc = stdout.channel.recv_exit_status()
            sout = [s.decode('utf-8') for s in stdout.read().splitlines()]
            if 'password' in sout[0].lower():
                sout.pop(0)
            if rc == 0:
                return (sout or True)
            else:
                return False
        proc = subprocess.Popen(cmd,
                                shell=True,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE
                                )
        stdout, stderr = proc.communicate()
        rc = proc.returncode
        if proc.returncode == 0:
            sout = stdout.splitlines()
            return ([s.decode('utf-8') for s in sout] or True)
        return False

    def setup(self):
        '''This MAY be used by a cluster to do prep work in case there are
        extra commands to be run even if a node lsit is given by the user, and
        thus get_nodes() would not be called
        '''
        pass

    def get_sos_prefix(self, facts):
        '''This wraps set_sos_prefix used by clusters.
        It is called by sosnode.finalize_sos_cmd() for each node
        '''
        try:
            return self.set_sos_prefix(facts)
        except:
            return ''

    def set_sos_prefix(self, facts):
        '''This may be overridden by clusters when needed.

        In a cluster this should return a string that is placed immediately
        before the 'sosreport' command, but will be after sudo if needed.

        If a cluster overrides this, it will need to be known if the the
        cluster needs to be sensitive to cluster nodes being Atomic Hosts.
        '''
        if facts['atomic']:
            cmd = 'atomic run --name=clustersos-tmp '
            img = self.config['image']
            return cmd + img

    def get_sos_path_strip(self, facts):
        '''This calls set_sos_path_strip that is used by clusters to determine
        if we need to remove a particular string from a returned sos path for
        any reason
        '''
        try:
            return self.set_sos_path_strip(facts)
        except:
            return ''

    def set_sos_path_strip(self, facts):
        '''This may be overriden by a cluster and used to set
        a string to be stripped from the return sos path if needed.

        For example, on Atomic Host, the sosreport gets written under
        /host/var/tmp in the container, but is available to scp under the
        standard /var/tmp after the container exits.

        If a cluster overrides this, it will need to be known if it needs to be
        sensitive to cluster nodes being Atomic Hosts.
        '''
        if facts['atomic']:
            return '/host'

    def get_cleanup_cmd(self, facts):
        '''This calls set_cleanup_cmd that is used by clusers to determine if
        clustersos needs to do additional cleanup on a node
        '''
        try:
            return self.set_cleanup_cmd(facts)
        except:
            return False

    def set_cleanup_cmd(self, facts):
        '''This should be overridden by a cluster and used to set an additional
        command to run during cleanup.

        The cluster should return a string containing the full cleanup
        command to run

        If a cluster overrides this, it will need to be known if the the
        cluster needs to be sensitive to cluster nodes being Atomic Hosts.
        '''
        if facts['atomic']:
            return 'docker rm clustersos-tmp'

    def check_enabled(self):
        '''This may be overridden by clusters

        This is called by clustersos on each cluster type that exists, and is
        meant to return True when the cluster type matches a criteria
        that indicates that is the cluster type is in use.

        Only the first cluster type to determine a match is run
        '''
        for pkg in self.packages:
            if self.is_installed(pkg):
                return True
        return False

    def get_nodes(self):
        '''This MUST be overridden by a cluster.
        A cluster should use this method to return a list or string that
        contains all the nodes that a report should be collected from
        '''
        pass

    def _get_nodes(self):
        try:
            return self.format_node_list()
        except Exception as e:
            print('Failed to get node list: %s' % e)
            return []

    def modify_sos_cmd(self):
        '''This is used to modify the sosreport command run on the nodes.
        By default, sosreport is run without any options, using this will
        allow the profile to specify what plugins to run or not and what
        options to use.

        This will NOT override user supplied options.
        '''

        if self.sos_plugins:
            for plug in self.sos_plugins:
                if plug not in self.config['sos_cmd']:
                    self.config['enable_plugins'].append(plug)
        if self.sos_options:
            for opt in self.sos_options:
                if opt not in self.config['sos_cmd']:
                    option = '%s=%s' % (opt, self.sos_options[opt])
                    self.config['plugin_option'].append(option)

    def format_node_list(self):
        '''Format the returned list of nodes from a cluster into a known
        format. This being a list that contains no duplicates
        '''
        try:
            nodes = self.get_nodes()
            if isinstance(nodes, list):
                node_list = list(set(nodes))
            if isinstance(nodes, str):
                node_list = [n.split(',').strip() for n in nodes]
                node_list = list(set(nodes))
            return node_list
        except Exception as e:
            print('Failed to format node list: %s' % e)
